# Copyright 2014 Konrad Podloucky, Michal Walkowski
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Mixin to add the capability of testing angular.js apps with selenium
webdrivers.
"""

from functools import wraps
from math import floor
from future.moves.urllib.parse import urljoin

from selenium.common.exceptions import NoSuchElementException
from selenium.webdriver.remote.command import Command
from selenium.webdriver.support.wait import WebDriverWait

from pkg_resources import resource_string

from .exceptions import AngularNotFoundException

CLIENT_SCRIPTS_DIR = 'protractor/extracted'
DEFER_LABEL = 'NG_DEFER_BOOTSTRAP!'
# These are commands that need synchronization with the angular app.
COMMANDS_NEEDING_WAIT = [
    Command.CLICK_ELEMENT,
    Command.SEND_KEYS_TO_ELEMENT,
    Command.GET_ELEMENT_TAG_NAME,
    Command.GET_ELEMENT_VALUE_OF_CSS_PROPERTY,
    Command.GET_ELEMENT_ATTRIBUTE,
    Command.GET_ELEMENT_TEXT,
    Command.GET_ELEMENT_SIZE,
    Command.GET_ELEMENT_LOCATION,
    Command.IS_ELEMENT_ENABLED,
    Command.IS_ELEMENT_SELECTED,
    Command.IS_ELEMENT_DISPLAYED,
    Command.SUBMIT_ELEMENT,
    Command.CLEAR_ELEMENT
]


def angular_wait_required(wrapped):
    @wraps(wrapped)
    def wait_for_angular(driver, *args, **kwargs):
        driver.wait_for_angular()
        return wrapped(driver, *args, **kwargs)
    return wait_for_angular


class WebDriverMixin(object):
    """A mixin for Selenium Webdrivers."""
    ignore_synchronization = False
    """If True, pytractor will not attempt to synchronize with the page before
    performing actions. This can be harmful because pytractor will not wait
    until $timeouts and $http calls have been processed, which can cause
    tests to become flaky. This should be used only when necessary, such as
    when a page continuously polls an API using $timeout.
    """  # docstring adapted from protractor.js

    def __init__(self, base_url='', root_element='body', script_timeout=10,
                 test_timeout=10, *args, **kwargs):
        self._base_url = base_url
        self._root_element = root_element
        self._test_timeout = test_timeout
        super(WebDriverMixin, self).__init__(*args, **kwargs)
        self.set_script_timeout(script_timeout)

    def _execute_client_script(self, script_name, *args, **kwargs):
        is_async = kwargs.pop('is_async', True)
        file_name = '{}.js'.format(script_name)
        js_script = resource_string(__name__,
                                    '{}/{}'.format(CLIENT_SCRIPTS_DIR,
                                                   file_name))
        if js_script:
            js_script = js_script.decode('UTF-8')
        if is_async:
            result = self.execute_async_script(js_script, *args)
        else:
            result = self.execute_script(js_script, *args)
        return result

    def wait_for_angular_enabled(self, enabled: bool = True):
        self.ignore_synchronization = not enabled

    def wait_for_angular(self):
        if self.ignore_synchronization:
            return
        else:
            return self._execute_client_script('waitForAngular',
                                               self._root_element,
                                               is_async=True)

    def execute(self, driver_command, params=None):
        # We also get called from WebElement methods/properties.
        if driver_command in COMMANDS_NEEDING_WAIT:
            self.wait_for_angular()

        return super(WebDriverMixin, self).execute(driver_command,
                                                   params=params)

    def _test_for_angular(self):
        return self._execute_client_script('testForAngular',
                                           floor(self._test_timeout / 1000))

    def _location_equals(self, location):
        result = self.execute_script('return window.location.href')
        return result == location

    @property
    @angular_wait_required
    def current_url(self):
        return super(WebDriverMixin, self).current_url

    @property
    @angular_wait_required
    def page_source(self):
        return super(WebDriverMixin, self).page_source

    @property
    @angular_wait_required
    def title(self):
        return super(WebDriverMixin, self).title

    @property
    @angular_wait_required
    def location_abs_url(self):
        return self._execute_client_script('getLocationAbsUrl',
                                           self._root_element, is_async=False)

    @angular_wait_required
    def find_elements_by_repeater(self, descriptor, using=None):
        return self._execute_client_script('findAllRepeaterRows',
                                           descriptor, False, using,
                                           is_async=False)

    @angular_wait_required
    def find_element(self, *args, **kwargs):
        return super(WebDriverMixin, self).find_element(*args, **kwargs)

    @angular_wait_required
    def find_elements(self, *args, **kwargs):
        return super(WebDriverMixin, self).find_elements(*args, **kwargs)

    @angular_wait_required
    def find_elements_by_binding(self, descriptor, using=None):
        elements = self._execute_client_script('findBindings', descriptor,
                                               False, using, is_async=False)
        return elements

    def find_element_by_binding(self, descriptor, using=None):
        elements = self.find_elements_by_binding(descriptor, using=using)
        if len(elements) == 0:
            raise NoSuchElementException(
                "No element found for binding descriptor"
                " '{}'".format(descriptor)
            )
        else:
            return elements[0]

    def find_element_by_exact_binding(self, descriptor, using=None):
        elements = self.find_elements_by_exact_binding(descriptor, using=using)
        if len(elements) == 0:
            raise NoSuchElementException(
                "No element found for binding descriptor"
                " '{}'".format(descriptor)
            )
        else:
            return elements[0]

    @angular_wait_required
    def find_elements_by_exact_binding(self, descriptor, using=None):
        elements = self._execute_client_script('findBindings', descriptor,
                                               True, using, is_async=False)
        return elements

    def find_element_by_model(self, descriptor, using=None):
        elements = self.find_elements_by_model(descriptor, using=using)
        if len(elements) == 0:
            raise NoSuchElementException(
                "No element found for model descriptor"
                " {}".format(descriptor)
            )
        else:
            return elements[0]

    @angular_wait_required
    def find_elements_by_model(self, descriptor, using=None):
        elements = self._execute_client_script('findByModel', descriptor,
                                               using, is_async=False)
        # Workaround for issue #10: findByModel.js returns None instead of empty
        # list if no element has been found.
        if elements is None:
            elements = []
        return elements

    def get(self, url):
        super(WebDriverMixin, self).get('about:blank')
        full_url = urljoin(str(self._base_url), str(url))
        self.execute_script(
            """
            window.name = "{}" + window.name;
            window.location.replace("{}");
            """.format(DEFER_LABEL, full_url)
        )
        wait = WebDriverWait(self, self._test_timeout)
        wait.until_not(self._location_equals, 'about:blank')

        if not self.ignore_synchronization:
            test_result = self._test_for_angular()
            angular_on_page = test_result[0]
            if not angular_on_page:
                message = test_result[1]
                raise AngularNotFoundException(
                    'Angular could not be found on page: {}:'
                    ' {}'.format(full_url, message)
                )
            # TODO: inject scripts here
            # return self.execute_script(
            #     'angular.resumeBootstrap(arguments[0]);'
            # )
            self.execute_script('angular.resumeBootstrap();')

    def refresh(self):
        url = self.execute_script('return window.location.href')
        self.get(url)

    @angular_wait_required
    def set_location(self, url):
        result = self._execute_client_script('setLocation', self._root_element,
                                             url, is_async=False)
        return result



# Copyright 2014 Konrad Podloucky
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Unit tests for pytractor.webdriver
"""
from importlib import import_module
import unittest

from selenium import webdriver

from pytractor.mixins import WebDriverMixin


class WebDriverModuleTest(unittest.TestCase):
    # All known webdrivers
    WEBDRIVERS = ['Android', 'Chrome', 'Firefox', 'Ie', 'Opera', 'PhantomJS',
                  'Remote', 'Safari']

    def setUp(self):
        self.module = import_module('pytractor.webdriver')

    def test_module_exports_all_webdrivers(self):
        for driver_name in self.WEBDRIVERS:
            self.assertTrue(
                hasattr(self.module, driver_name),
                '{} was not found in the module!'.format(driver_name)
            )

    def test_exported_webdrivers_are_classes(self):
        for driver_name in self.WEBDRIVERS:
            klass = getattr(self.module, driver_name)
            self.assertIsInstance(
                klass, type, '{} is not a class!'.format(driver_name)
            )

    def test_exported_webdrivers_are_subclasses_of_webdriver_and_mixin(self):
        for driver_name in self.WEBDRIVERS:
            selenium_driver = getattr(webdriver, driver_name)
            pytractor_driver = getattr(self.module, driver_name)
            self.assertTrue(issubclass(pytractor_driver, selenium_driver),
                            'pytractor.webdriver.{0} is not a subclass of '
                            'selenium.webdriver.{0}'.format(driver_name))
            self.assertTrue(issubclass(pytractor_driver, WebDriverMixin),
                            'pytractor.webdriver.{} is not a subclass of '
                            'pytractor.mixins.WebDriverMixin'
                            .format(driver_name))



# Copyright 2014 Konrad Podloucky
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This is the web server that serves the angular app that we use for testing.
It is started by setup_package() in __init__.py
"""

import logging
import os
import signal
import http.server
import socketserver
import time

PORT = 8000

logger = logging.getLogger(__name__)


class TestServerHandler(http.server.SimpleHTTPRequestHandler):
    """
    The handler for the web server. It has the same functionality as the
    server for protractor's test app.
    """
    def send_text_response(self, text):
        self.send_response(200)
        self.send_header("Content-type", 'text/plain')
        self.send_header("Content-Length", len(text))
        self.end_headers()
        self.wfile.write(text.encode('utf-8'))

    def do_GET(self):
        if self.path == '/fastcall':
            self.send_text_response('done')
        elif self.path == '/slowcall':
            time.sleep(5)
            self.send_text_response('finally done')
        elif self.path == '/fastTemplateUrl':
            self.send_text_response('fast template contents')
        elif self.path == '/slowTemplateUrl':
            time.sleep(5)
            self.send_text_response('slow template contents')
        else:
            http.server.SimpleHTTPRequestHandler.do_GET(self)

    def log_message(self, msg_format, *args):
        """Use python logging to avoid lots of output during testing."""
        logger.info("TESTSERVER: %s - - [%s] %s\n" %
                    (self.client_address[0],
                     self.log_date_time_string(),
                     msg_format % args))


class SimpleWebServerProcess(object):
    """
    A simple webserver for serving pages for testing.
    """
    HOST = 'localhost'
    PORT = 9999
    APP_DIR = 'testapp'
    _pid = None

    def run(self):
        self._pid = os.fork()
        if self._pid == 0:
            self.start_server()
        else:
            logger.debug('Started webserver child as pid {} on'
                         ' port {}'.format(self._pid, self.PORT))
            # wait 5 seconds for server to start
            time.sleep(5)

    def start_server(self):
        module_path = __file__
        server_path = os.path.join(os.path.dirname(module_path), self.APP_DIR)
        logger.debug('Starting webserver for path {} on'
                     ' port {}'.format(server_path, self.PORT))
        os.chdir(server_path)
        handler = TestServerHandler
        socketserver.TCPServer.allow_reuse_address = True
        httpd = socketserver.TCPServer((self.HOST, self.PORT), handler)
        httpd.serve_forever()

    def stop(self):
        if self._pid != 0:
            logger.debug('Sending SIGTERM to webserver child with'
                         ' pid {}'.format(self._pid))
            os.kill(self._pid, signal.SIGTERM)
            os.waitpid(self._pid, 0)

if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG)
    # start blocking server, do not fork into the background
    process = SimpleWebServerProcess()
    process.start_server()



# Copyright 2014 Konrad Podloucky
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import os.path
import signal
import http.server
import socketserver
import time

from .testserver import SimpleWebServerProcess

WEBSERVER_PROCESS = None


def setup_package():
    global WEBSERVER_PROCESS  # pylint: disable=global-statement
    WEBSERVER_PROCESS = SimpleWebServerProcess()
    WEBSERVER_PROCESS.run()


def teardown_package():
    WEBSERVER_PROCESS.stop()



# Copyright 2014 Konrad Podloucky
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from mock import MagicMock, patch, PropertyMock, DEFAULT

from selenium.common.exceptions import NoSuchElementException

from pytractor.exceptions import AngularNotFoundException
from pytractor.mixins import (WebDriverMixin, angular_wait_required,
                              CLIENT_SCRIPTS_DIR)


try:  # FIXME: find another way
    import builtins
    super_str = 'builtins.super'
except ImportError:
    super_str = 'builtins.super'


class AngularWaitRequiredDecoratorTest(unittest.TestCase):
    @angular_wait_required
    def wrapped_function(self, *args, **kwargs):
        return self.check_function(*args, **kwargs)

    def check_function(self, *args, **kwargs):
        pass

    def test_angular_wait_required(self):
        with patch.multiple(self, wait_for_angular=DEFAULT,
                            check_function=DEFAULT, create=True) as patched:
            mock_wait_for_angular = patched['wait_for_angular']
            mock_check_function = patched['check_function']
            mock_arg = MagicMock()
            mock_kwarg = MagicMock()
            result = self.wrapped_function(mock_arg, kwarg=mock_kwarg)
        # result should be the result of the wrapped function
        self.assertIs(result, mock_check_function.return_value)
        # wait_for_angular() should have been called
        mock_wait_for_angular.assert_called_once_with()
        # the check function should have been called
        mock_check_function.assert_called_once_with(mock_arg,
                                                    kwarg=mock_kwarg)


class WebDriverMixinConstructorTest(unittest.TestCase):
    class ConstructorTester(object):
        """Class for checking calls to __init__"""
        def __init__(self, *args, **kwargs):
            self.constructor_called(*args, **kwargs)

        def constructor_called(self, *args, **kwargs):
            """
            This function will be called by __init__. We can hook onto this
            to check for __init__ calls.
            """
            pass

    class TestDriver(WebDriverMixin, ConstructorTester):
        pass

    def test_constructor(self):
        base_url = 'BASEURL'
        root_element = 'ROOTEL'
        test_timeout = 'TESTTIMEOUT'
        script_timeout = 'SCRIPTTIMEOUT'

        with patch.object(
            self.ConstructorTester, 'constructor_called'
        ) as mock_constructor_called, patch.object(
            self.TestDriver, 'set_script_timeout', create=True
        ) as mock_set_script_timeout:
            instance = self.TestDriver(base_url, root_element,
                                       test_timeout=test_timeout,
                                       script_timeout=script_timeout)

        # verify that properties have been set
        self.assertIs(instance._base_url, base_url)
        self.assertIs(instance._root_element, root_element)
        self.assertIs(instance._test_timeout, test_timeout)
        # verify that the super class' constructor has been called
        mock_constructor_called.assert_called_once_with()
        # verify that set_script_timeout has been called
        mock_set_script_timeout.assert_called_once_with(script_timeout)


class WebDriverMixinTest(unittest.TestCase):
    def setUp(self):
        set_script_timeout_patcher = patch.object(
            WebDriverMixin, 'set_script_timeout', create=True
        )
        self.mock_set_script_timeout = set_script_timeout_patcher.start()
        self.addCleanup(set_script_timeout_patcher.stop)
        self.mock_root_element = MagicMock()
        self.instance = WebDriverMixin('http://localhost',
                                       self.mock_root_element)

    @patch('pytractor.mixins.resource_string')
    def verify__execute_client_script_call(self, is_async, mock_resource_string):
        with patch.multiple(
            self.instance,
            execute_async_script=DEFAULT, execute_script=DEFAULT,
            create=True
        ) as mock_execute:
            (mock_execute_script,
             mock_execute_async_script) = [mock_execute.get(func_name)
                                           for func_name in
                                           ('execute_script',
                                            'execute_async_script')]
            mock_arg = MagicMock()
            result = self.instance._execute_client_script('SCRIPT', mock_arg,
                                                          is_async=is_async)
        # the script was read correctly with resource_string()
        mock_resource_string.assert_called_once_with(
            'pytractor.mixins',
            '{}/{}.js'.format(CLIENT_SCRIPTS_DIR, 'SCRIPT')
        )
        # execute_async_script or execute_script were called (but not both)
        script_content = mock_resource_string.return_value.decode()
        if is_async:
            mock_execute_async_script.assert_called_once_with(script_content,
                                                              mock_arg)
            self.assertEqual(len(mock_execute_script.mock_calls), 0)
            # the result is the one from execute_async_script()
            self.assertIs(result, mock_execute_async_script.return_value)
        else:
            mock_execute_script.assert_called_once_with(script_content,
                                                        mock_arg)
            self.assertEqual(len(mock_execute_async_script.mock_calls), 0)
            # the result is the one from execute_script()
            self.assertIs(result, mock_execute_script.return_value)

    def test__execute_client_script_async(self):
        self.verify__execute_client_script_call(True)

    def test__execute_client_script_sync(self):
        self.verify__execute_client_script_call(False)

    def verify_function_executes_script_with(self, func_to_call,
                                             script_name, *script_args,
                                             **script_kwargs):
        with patch.object(
            self.instance, '_execute_client_script'
        ) as mock_execute_client_script:
            result = func_to_call()
        mock_execute_client_script.assert_called_once_with(
            script_name,
            *script_args, **script_kwargs
        )
        self.assertIs(result, mock_execute_client_script.return_value)

    def test_wait_for_angular(self):
        self.verify_function_executes_script_with(
            self.instance.wait_for_angular,
            'waitForAngular', self.mock_root_element, is_async=True
        )

    def test_wait_for_angular_does_not_call_script_if_ignore_synchronization(
        self
    ):
        """wait_for_angular() must not call the waitForAngular script, if
        ignore_synchronization is set to True."""
        self.instance.ignore_synchronization = True

        with patch.object(
            self.instance, '_execute_client_script'
        ) as mock_execute_client_script:

            self.instance.wait_for_angular()

        self.assertEqual(mock_execute_client_script.call_count, 0)

    def test__test_for_angular(self):
        self.instance._test_timeout = 5000
        self.verify_function_executes_script_with(
            self.instance._test_for_angular,
            'testForAngular', self.instance._test_timeout / 1000
        )

    def test__location_equals(self):
        with patch.object(
            self.instance, 'execute_script', create=True
        ) as mock_execute_script:
            mock_location = MagicMock()
            mock_execute_script.return_value = MagicMock(__eq__=MagicMock())
            result = self.instance._location_equals(mock_location)
        mock_execute_script.assert_called_once_with(
            'return window.location.href'
        )
        script_result = mock_execute_script.return_value
        script_result.__eq__.assert_called_once_with(mock_location)
        self.assertIs(result, script_result.__eq__.return_value)

    # The following tests test some properties that use the wait_for_angular
    # decorator and fetch the property from the super class.
    def verify_super_property_called_with_wait(self, prop_name):
        """
        Verifies that accessing the given property calls the equally
        named property on the super class.
        """
        with patch(
                super_str
        ) as mock_super, patch.object(
            self.instance, 'wait_for_angular'
        ) as mock_wait_for_angular:
            # setup the mocked property
            mock_prop = PropertyMock(name='super.{}'.format(prop_name))
            setattr(type(mock_super.return_value), prop_name, mock_prop)

            result = getattr(self.instance, prop_name)

        mock_wait_for_angular.assert_called_once_with()
        mock_super.assert_called_once_with(WebDriverMixin, self.instance)
        mock_prop.assert_called_once_with()
        self.assertIs(result, mock_prop.return_value)

    def test_current_url(self):
        self.verify_super_property_called_with_wait('current_url')

    def test_page_source(self):
        self.verify_super_property_called_with_wait('page_source')

    def test_title(self):
        self.verify_super_property_called_with_wait('title')

    # Tests for methods that delegate to the super method
    def verify_super_method_called_with_wait(self, method_name):
        """
        Verifies that calling the given method calls the equally
        named method on the super class.
        """
        mock_args = [MagicMock(), MagicMock()]
        with patch(
            super_str
        ) as mock_super, patch.object(
            self.instance, 'wait_for_angular'
        ) as mock_wait_for_angular:
            mock_super_method = getattr(mock_super.return_value, method_name)
            method = getattr(self.instance, method_name)
            result = method(*mock_args)

        mock_wait_for_angular.assert_called_once_with()
        mock_super.assert_called_once_with(WebDriverMixin, self.instance)
        mock_super_method.assert_called_once_with(*mock_args)
        self.assertIs(result, mock_super_method.return_value)

    def test_find_element(self):
        self.verify_super_method_called_with_wait('find_element')

    def test_find_elements(self):
        self.verify_super_method_called_with_wait('find_elements')

    # tests for other methods
    def test_find_elements_by_binding(self):
        mock_descriptor = MagicMock()
        mock_using = MagicMock()
        with patch.multiple(
            self.instance, wait_for_angular=DEFAULT,
            _execute_client_script=DEFAULT
        ) as mock_methods:
            result = self.instance.find_elements_by_binding(mock_descriptor,
                                                            mock_using)
        mock_methods['wait_for_angular'].assert_called_once_with()
        mock_methods['_execute_client_script'].assert_called_once_with(
            'findBindings', mock_descriptor, False, mock_using, is_async=False
        )
        self.assertIs(result,
                      mock_methods['_execute_client_script'].return_value)

    def test_find_element_by_binding_no_element(self):
        mock_descriptor = MagicMock()
        mock_using = MagicMock()
        with patch.object(
            self.instance, 'find_elements_by_binding'
        ) as mock_find_elements_by_binding:
            mock_find_elements_by_binding.return_value = []
            with self.assertRaises(NoSuchElementException):
                self.instance.find_element_by_binding(mock_descriptor,
                                                      mock_using)
        mock_find_elements_by_binding.assert_called_once_with(
            mock_descriptor, using=mock_using
        )

    def test_find_element_by_binding_with_element(self):
        mock_descriptor = MagicMock()
        mock_using = MagicMock()
        mock_element = MagicMock()
        with patch.object(
            self.instance, 'find_elements_by_binding'
        ) as mock_find_elements_by_binding:
            mock_find_elements_by_binding.return_value = [mock_element]
            result = self.instance.find_element_by_binding(mock_descriptor,
                                                           mock_using)
        mock_find_elements_by_binding.assert_called_once_with(
            mock_descriptor, using=mock_using
        )
        self.assertIs(result, mock_element)

    def test_get_with_angular(self):
        mock_url = MagicMock()
        with patch(
            super_str
        ) as mock_super, patch(
            'pytractor.mixins.WebDriverWait'
        ) as mock_webdriverwait_class, patch.multiple(
            self.instance, _test_for_angular=DEFAULT, execute_script=DEFAULT,
            create=True  # needed for execute_script
        ) as mock_methods:
            mock_test_for_angular = mock_methods['_test_for_angular']
            # return a truthy value to indicate that angular was found
            mock_test_for_angular.return_value = (True,)
            mock_execute_script = mock_methods['execute_script']

            self.instance.get(mock_url)
        mock_super.assert_called_once_with(WebDriverMixin, self.instance)
        mock_super.return_value.get.assert_called_once_with('about:blank')
        self.assertEqual(len(mock_execute_script.mock_calls), 2)
        mock_webdriverwait_class.assert_called_once_with(self.instance,
                                                         10)
        mock_webdriverwait_instance = mock_webdriverwait_class.return_value
        mock_webdriverwait_instance.until_not.assert_called_once_with(
            self.instance._location_equals, 'about:blank'
        )
        mock_test_for_angular.assert_called_once_with()

    def test_get_without_angular(self):
        mock_url = MagicMock()
        with patch(
            super_str
        ) as mock_super, patch(
            'pytractor.mixins.WebDriverWait'
        ) as mock_webdriverwait_class, patch.multiple(
            self.instance, _test_for_angular=DEFAULT, execute_script=DEFAULT,
            create=True  # needed for execute_script
        ) as mock_methods:
            mock_test_for_angular = mock_methods['_test_for_angular']
            # return a falsy value to indicate that angular was not found
            mock_test_for_angular.return_value = (False, 'ERROR')
            mock_execute_script = mock_methods['execute_script']

            with self.assertRaises(AngularNotFoundException):
                self.instance.get(mock_url)
        mock_super.assert_called_once_with(WebDriverMixin, self.instance)
        mock_super.return_value.get.assert_called_once_with('about:blank')
        self.assertEqual(len(mock_execute_script.mock_calls), 1)
        mock_webdriverwait_class.assert_called_once_with(self.instance,
                                                         10)
        mock_webdriverwait_instance = mock_webdriverwait_class.return_value
        mock_webdriverwait_instance.until_not.assert_called_once_with(
            self.instance._location_equals, 'about:blank'
        )
        mock_test_for_angular.assert_called_once_with()

    def test_get_does_not_test_for_angular_if_ignore_synchronization_is_true(
        self
    ):
        """Verify that get() does not call _test_for_angular if
        ignore_synchronization is set to True."""
        mock_url = MagicMock()
        with patch(
            super_str
        ) as mock_super, patch(
            'pytractor.mixins.WebDriverWait'
        ) as mock_webdriverwait_class, patch.multiple(
            self.instance, _test_for_angular=DEFAULT, execute_script=DEFAULT,
            create=True  # needed for execute_script
        ) as mock_methods:
            mock_test_for_angular = mock_methods['_test_for_angular']

            self.instance.ignore_synchronization = True
            self.instance.get(mock_url)

        self.assertEqual(mock_test_for_angular.call_count, 0)

    def test_refresh(self):
        with patch.multiple(
            self.instance, get=DEFAULT, execute_script=DEFAULT,
            create=True  # needed for execute_script()
        ) as mock_methods:
            self.instance.refresh()
        mock_execute_script, mock_get = (mock_methods['execute_script'],
                                         mock_methods['get'])
        self.assertEqual(mock_execute_script.call_count, 1)
        mock_get.assert_called_once_with(mock_execute_script.return_value)

    def test_find_element_by_exact_binding_calls_find_elements(self):
        mock_descriptor = MagicMock()
        mock_using = MagicMock()
        mock_element = MagicMock()
        with patch.object(
            self.instance, 'find_elements_by_exact_binding',
            return_value=[mock_element]
        ) as mock_find_elements_by_exact_binding:
            result = self.instance.find_element_by_exact_binding(
                mock_descriptor, mock_using
            )
        mock_find_elements_by_exact_binding.assert_called_once_with(
            mock_descriptor, using=mock_using
        )
        self.assertIs(result, mock_element)

    def test_find_element_by_exact_binding_raises_error_if_nothing_found(self):
        mock_descriptor = MagicMock()
        mock_using = MagicMock()
        with patch.object(
            self.instance, 'find_elements_by_exact_binding', return_value=[]
        ) as mock_find_elements_by_exact_binding:
            with self.assertRaises(NoSuchElementException):
                self.instance.find_element_by_exact_binding(
                    mock_descriptor, mock_using
                )
        mock_find_elements_by_exact_binding.assert_called_once_with(
            mock_descriptor, using=mock_using
        )

    def test_find_elements_by_exact_binding_calls_protractor_script(self):
        mock_descriptor = MagicMock()
        mock_using = MagicMock()

        with patch.multiple(
            self.instance, wait_for_angular=DEFAULT,
            _execute_client_script=DEFAULT
        ) as mock_methods:
            result = self.instance.find_elements_by_exact_binding(
                mock_descriptor, mock_using
            )

        mock_methods['wait_for_angular'].assert_called_once_with()
        mock_methods['_execute_client_script'].assert_called_once_with(
            'findBindings', mock_descriptor, True, mock_using, is_async=False
        )
        self.assertIs(result,
                      mock_methods['_execute_client_script'].return_value)

    def test_find_element_by_model_calls_find_elements(self):
        mock_descriptor = MagicMock()
        mock_using = MagicMock()
        with patch.object(
            self.instance, 'find_elements_by_model', return_value=[]
        ) as mock_find_elements_by_model:
            with self.assertRaises(NoSuchElementException):
                self.instance.find_element_by_model(
                    mock_descriptor, mock_using
                )
        mock_find_elements_by_model.assert_called_once_with(
            mock_descriptor, using=mock_using
        )

    def test_find_element_by_model_raises_error_if_nothing_found(self):
        mock_descriptor = MagicMock()
        mock_using = MagicMock()
        with patch.object(
            self.instance, 'find_elements_by_model', return_value=[]
        ) as mock_find_elements_by_model:
            with self.assertRaises(NoSuchElementException):
                self.instance.find_element_by_model(
                    mock_descriptor, mock_using
                )
        mock_find_elements_by_model.assert_called_once_with(
            mock_descriptor, using=mock_using
        )

    def test_find_elements_by_model_calls_protractor_script(self):
        mock_descriptor = MagicMock()
        mock_using = MagicMock()

        with patch.multiple(
            self.instance, wait_for_angular=DEFAULT,
            _execute_client_script=DEFAULT
        ) as mock_methods:
            result = self.instance.find_elements_by_model(
                mock_descriptor, mock_using
            )

        mock_methods['wait_for_angular'].assert_called_once_with()
        mock_methods['_execute_client_script'].assert_called_once_with(
            'findByModel', mock_descriptor, mock_using, is_async=False
        )
        self.assertIs(result,
                      mock_methods['_execute_client_script'].return_value)

    def test_find_elements_by_model_returns_empty_list_if_script_returns_none(
        self
    ):
        """Verify that find_elements_by_model() returns an empty list if
        the protractor script returns None.
        This is a test for issue #10
        """
        with patch.object(self.instance, '_execute_client_script',
                          return_value=None):
            result = self.instance.find_elements_by_model('does-not-exist')

        self.assertIsInstance(result, list)
        self.assertEqual(len(result), 0)

    def test_location_abs_url_calls_protractor_script(self):
        with patch.multiple(
            self.instance, wait_for_angular=DEFAULT,
            _execute_client_script=DEFAULT
        ) as mock_methods:
            result = self.instance.location_abs_url

        mock_methods['wait_for_angular'].assert_called_once_with()
        mock_methods['_execute_client_script'].assert_called_once_with(
            'getLocationAbsUrl', self.instance._root_element, is_async=False
        )
        self.assertIs(result,
                      mock_methods['_execute_client_script'].return_value)

    def test_set_location_calls_protractor_script(self):
        url = 'http://a.new.locat.ion/'
        with patch.multiple(
            self.instance, wait_for_angular=DEFAULT,
            _execute_client_script=DEFAULT
        ) as mock_methods:
            result = self.instance.set_location(url)

        mock_methods['wait_for_angular'].assert_called_once_with()
        mock_methods['_execute_client_script'].assert_called_once_with(
            'setLocation', self.instance._root_element, url, is_async=False
        )
        self.assertIs(result,
                      mock_methods['_execute_client_script'].return_value)



# Copyright 2014 Konrad Podloucky
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Simple tests for Firefox and Chrome drivers. We just see if the drivers can
be instantiated and run a simple test.
"""
import unittest

from selenium.webdriver.remote.webelement import WebElement

# pylint: disable=no-name-in-module
from pytractor.webdriver import Firefox, Chrome
# pylint: enable=no-name-in-module

from . import SimpleWebServerProcess


class WebDriverTestBase(object):

    @classmethod
    def setUpClass(cls):
        cls.driver = cls.driver_class(
            'http://localhost:{}/'.format(SimpleWebServerProcess.PORT),
            'body'
        )

    @classmethod
    def tearDownClass(cls):
        cls.driver.quit()

    def test_find_element_by_binding(self):
        self.driver.get('index.html#/form')
        element = self.driver.find_element_by_binding('greeting')
        self.assertIsInstance(element, WebElement)
        self.assertEqual(element.text, 'Hiya')


class FirefoxWebDriverTest(WebDriverTestBase, unittest.TestCase):
    driver_class = Firefox


class ChromeWebDriverTest(WebDriverTestBase, unittest.TestCase):
    driver_class = Chrome



# encoding: utf-8
# Copyright 2014 Konrad Podloucky, Michal Walkowski
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import TestCase

from selenium.common.exceptions import NoSuchElementException
from selenium.webdriver.remote.webelement import WebElement

from .testdriver import TestDriver
from .testserver import SimpleWebServerProcess


class LocatorTestCase(TestCase):
    """Test case class for testing locators."""

    @classmethod
    def setUpClass(cls):
        cls.driver = TestDriver(
            'http://localhost:{}/'.format(SimpleWebServerProcess.PORT),
            'body'
        )

    @classmethod
    def tearDownClass(cls):
        cls.driver.quit()


class ByBindingLocatorTest(LocatorTestCase):
    """Tests the locators of the WebDriverMixin that deal with bindings."""

    def setUp(self):
        self.driver.get('index.html#/form')


    def test_find_element_by_binding_raises_error_if_no_element_matches(self):
        with self.assertRaises(NoSuchElementException):
            self.driver.find_element_by_binding('no-such-binding')

    def test_find_element_by_binding_returns_correct_element(self):
        element = self.driver.find_element_by_binding('greeting')
        self.assertIsInstance(element, WebElement)
        self.assertEqual(element.text, 'Hiya')

    def test_find_element_by_binding_finds_element_by_partial_name(self):
        element = self.driver.find_element_by_binding('greet')
        self.assertIsInstance(element, WebElement)
        self.assertEqual(element.text, 'Hiya')

    def test_find_element_by_binding_finds_element_with_ng_bind(self):
        element = self.driver.find_element_by_binding('username')
        self.assertIsInstance(element, WebElement)
        self.assertEqual(element.text, 'Anon')

    def test_find_element_by_binding_finds_element_with_ng_bind_template(self):
        element = self.driver.find_element_by_binding('nickname|uppercase')
        self.assertIsInstance(element, WebElement)
        self.assertEqual(element.text, '(ANNIE)')

    def test_find_element_by_exact_binding_finds_correct_element(self):
        element = self.driver.find_element_by_exact_binding('greeting')
        self.assertIsInstance(element, WebElement)
        self.assertEqual(element.text, 'Hiya')

    def test_find_element_by_exact_binding_needs_complete_binding_name(self):
        with self.assertRaises(NoSuchElementException):
            self.driver.find_element_by_exact_binding('greet')


class ByModelLocatorTest(LocatorTestCase):
    def setUp(self):
        self.driver.get('index.html#/form')

    def test_find_element_by_model_finds_element_by_text_input_model(self):
        username = self.driver.find_element_by_model('username')
        name = self.driver.find_element_by_binding('username')

        username.clear()
        self.assertEqual(name.text, '')

        username.send_keys('Jane Doe')
        self.assertEqual(name.text, 'Jane Doe')

        username.clear()
        self.assertEqual(name.text, '')

    def test_find_element_by_model_finds_element_by_checkbox_input_model(self):
        shower = self.driver.find_element_by_id('shower')
        self.assertTrue(shower.is_displayed())

        colors = self.driver.find_element_by_model('show')
        colors.click()

        shower = self.driver.find_element_by_id('shower')
        self.assertFalse(shower.is_displayed())

    def test_find_element_by_model_finds_textarea_by_model(self):
        about = self.driver.find_element_by_model('aboutbox')
        self.assertEqual(about.get_attribute('value'), 'This is a text box')

        about.clear()
        about.send_keys('Something else to write about')

        self.assertEqual(about.get_attribute('value'),
                         'Something else to write about')

    def test_find_elements_by_model_find_multiple_selects_by_model(self):
        selects = self.driver.find_elements_by_model('dayColor.color')

        self.assertEqual(len(selects), 3)

    def test_find_element_by_model_finds_the_selected_option(self):
        select = self.driver.find_element_by_model('fruit')
        selected_option = select.find_element_by_css_selector('option:checked')

        self.assertEqual(selected_option.text, 'apple')

    def test_find_element_by_model_finds_inputs_with_alternate_attribute_forms(
        self
    ):
        letter_list = self.driver.find_element_by_id('letterlist')
        self.assertEqual(letter_list.text, '')

        self.driver.find_element_by_model('check.w').click()
        self.assertEqual(letter_list.text, 'w')

        self.driver.find_element_by_model('check.x').click()
        self.assertEqual(letter_list.text, 'wx')

    def test_find_elements_by_model_finds_multiple_inputs(self):
        inputs = self.driver.find_elements_by_model('color')

        self.assertEqual(len(inputs), 3)

    def test_find_elements_by_model_returns_empty_list_if_nothing_found(self):
        """find_elements_by_model() should return an empty list if no elements
        have been found.

        This tests for issue #10 which comes from protractor's findByModel.js
        script.
        """
        result = self.driver.find_elements_by_model('this-model-does-not-exist')

        self.assertIsInstance(result, list)
        self.assertEqual(len(result), 0)


class ByRepeaterTestCase(LocatorTestCase):
    def setUp(self):
        self.driver.get('index.html#/repeater')

    def test_find_elements_by_repeater_returns_correct_element(self):
        element = self.driver.find_elements_by_repeater('allinfo in days')
        self.assertEqual(len(element), 5)
        self.assertEqual(element[0].text, 'M Monday')
        self.assertEqual(element[1].text, 'T Tuesday')
        self.assertEqual(element[2].text, 'W Wednesday')
        self.assertEqual(element[3].text, 'Th Thursday')
        self.assertEqual(element[4].text, 'F Friday')

    def test_find_elements_by_repeater_returns_empty_list(self):
        self.assertFalse(
            self.driver.find_elements_by_repeater('no-such in days')
        )



# Copyright 2014 Konrad Podloucky
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Selenium webdrivers with added angular.js awareness.
"""

from selenium import webdriver as selenium_webdriver
from selenium.webdriver.remote.webdriver import WebDriver

from .mixins import WebDriverMixin

module_dict = globals()  # pylint: disable=invalid-name
__all__ = []

# build classes derived from selenium webdrivers and our WebDriverMixin
for name in dir(selenium_webdriver):
    export = getattr(selenium_webdriver, name)
    if isinstance(export, type) and issubclass(export, WebDriver):
        module_dict[name] = type(name, (WebDriverMixin, export), {})
        __all__.append(name)

__all__ = tuple(__all__)



# encoding: utf-8
# Copyright 2014 Konrad Podloucky
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import TestCase

from pytractor.exceptions import AngularNotFoundException

from .testdriver import TestDriver
from .testserver import SimpleWebServerProcess


class WebDriverBaseTest(TestCase):
    """Tests the WebDriverMixin."""

    @classmethod
    def setUpClass(cls):
        cls.driver = TestDriver(
            'http://localhost:{}/'.format(SimpleWebServerProcess.PORT),
            'body'
        )

    @classmethod
    def tearDownClass(cls):
        cls.driver.quit()

    def test_get_no_angular(self):
        with self.assertRaises(AngularNotFoundException):
            self.driver.get('index-no-angular.html')

    def test_get_no_angular_does_not_fail_if_ignore_synchronization_set(self):
        self.driver.ignore_synchronization = True
        self.driver.get('index-no-angular.html')



# Copyright 2014 Konrad Podloucky
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Contains tests that wait for angular's processing to finish.
"""

from unittest import TestCase

from .testdriver import TestDriver
from .testserver import SimpleWebServerProcess


class AngularWaitTest(TestCase):
    """Test case class for testing waiting for angular."""

    @classmethod
    def setUpClass(cls):
        cls.driver = TestDriver(
            'http://localhost:{}/'.format(SimpleWebServerProcess.PORT),
            'body'
        )

    @classmethod
    def tearDownClass(cls):
        cls.driver.quit()

    def setUp(self):
        self.driver.get('index.html#/async')

    def test_waits_for_http_calls(self):
        status = self.driver.find_element_by_binding('slowHttpStatus')
        button = self.driver.find_element_by_css_selector(
            '[ng-click="slowHttp()"]'
        )
        self.assertEqual(status.text, 'not started')

        button.click()

        self.assertEqual(status.text, 'done')

    def test_waits_for_long_javascript_execution(self):
        status = self.driver.find_element_by_binding('slowFunctionStatus')
        button = self.driver.find_element_by_css_selector(
            '[ng-click="slowFunction()"]'
        )
        self.assertEqual(status.text, 'not started')

        button.click()

        self.assertEqual(status.text, 'done')

    def test_does_not_wait_for_timeout(self):
        status = self.driver.find_element_by_binding('slowTimeoutStatus')
        button = self.driver.find_element_by_css_selector(
            '[ng-click="slowTimeout()"]'
        )
        self.assertEqual(status.text, 'not started')

        button.click()

        self.assertEqual(status.text, 'pending...')

    def test_waits_for_timeout_service(self):
        status = self.driver.find_element_by_binding(
            'slowAngularTimeoutStatus'
        )
        button = self.driver.find_element_by_css_selector(
            '[ng-click="slowAngularTimeout()"]'
        )
        self.assertEqual(status.text, 'not started')

        button.click()

        self.assertEqual(status.text, 'done')

    def test_waits_for_timeout_service_then_a_promise(self):
        status = self.driver.find_element_by_binding(
            'slowAngularTimeoutPromiseStatus'
        )
        button = self.driver.find_element_by_css_selector(
            '[ng-click="slowAngularTimeoutPromise()"]'
        )
        self.assertEqual(status.text, 'not started')

        button.click()

        self.assertEqual(status.text, 'done')

    def test_waits_for_long_http_call_then_a_promise(self):
        status = self.driver.find_element_by_binding('slowHttpPromiseStatus')
        button = self.driver.find_element_by_css_selector(
            '[ng-click="slowHttpPromise()"]'
        )
        self.assertEqual(status.text, 'not started')

        button.click()

        self.assertEqual(status.text, 'done')

    def test_waits_for_slow_routing_changes(self):
        status = self.driver.find_element_by_binding('routingChangeStatus')
        button = self.driver.find_element_by_css_selector(
            '[ng-click="routingChange()"]'
        )
        self.assertEqual(status.text, 'not started')

        button.click()

        self.assertIn('polling mechanism', self.driver.page_source)

    def test_waits_for_slow_ng_include_templates_to_load(self):
        status = self.driver.find_element_by_css_selector('.included')
        button = self.driver.find_element_by_css_selector(
            '[ng-click="changeTemplateUrl()"]'
        )
        self.assertEqual(status.text, 'fast template contents')

        button.click()

        # need to refetch status as the element has been removed from the DOM
        status = self.driver.find_element_by_css_selector('.included')
        self.assertEqual(status.text, 'slow template contents')



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

# Copyright 2014 Konrad Podloucky
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Setuptools script for building pytractor.
"""

from setuptools import setup, find_packages
from codecs import open  # To use a consistent encoding
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the relevant file
with open(path.join(here, 'DESCRIPTION.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='pytractor-new',

    version='0.1.5',

    description='Selenium testing for Angular.js apps',
    long_description=long_description,

    url='https://github.com/kpodl/pytractor',

    author='Konrad Podloucky',
    author_email='konrad+pytractor@crunchy-frog.org',

    license='Apache 2.0',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 3 - Alpha',

        'Intended Audience :: Developers',
        'Topic :: Software Development :: Testing',

        'License :: OSI Approved :: Apache Software License',

        'Programming Language :: Python :: 2.7',
    ],

    keywords='selenium angular.js testing',

    package_dir={'': 'src'},

    packages=find_packages('src'),

    # For package data see MANIFEST.in
    include_package_data=True,

    install_requires=[
        'selenium>=2.48.0',
        'future>=0.15.2'
    ],

    tests_require=[
        'nose>=1.3.7',
        'mock>=1.3.0',
    ],
    test_suite='nose.collector',
    use_2to3=True
)



# Copyright 2014 Konrad Podloucky
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Contains tests that wait for angular's processing to finish.
"""

from unittest import TestCase

from .testdriver import TestDriver
from .testserver import SimpleWebServerProcess


class HelperFunctionTestCase(TestCase):
    """Tests for helper functions."""

    @classmethod
    def setUpClass(cls):
        cls.driver = TestDriver(
            'http://localhost:{}/'.format(SimpleWebServerProcess.PORT),
            'body'
        )

    @classmethod
    def tearDownClass(cls):
        cls.driver.quit()

    def setUp(self):
        self.driver.get('index.html')

    def test_location_abs_url_returns_absolute_url(self):
        url = self.driver.location_abs_url
        self.assertIn('/form', url)

        repeater_button = self.driver.find_element_by_link_text('repeater')
        repeater_button.click()
        url = self.driver.location_abs_url

        self.assertIn('/repeater', url)

    def test_set_location_navigates_to_another_url(self):
        self.driver.set_location('/repeater')
        url = self.driver.location_abs_url

        self.assertIn('/repeater', url)
